@@ -17,7 +17,7 @@ except ImportError:
 from redis._compat import (xrange, imap, byte_to_chr, unicode, long,
                            nativestr, basestring, iteritems,
                            LifoQueue, Empty, Full, urlparse, parse_qs,
-                           recv, recv_into, select, unquote)
+                           recv, recv_into, unquote)
 from redis.exceptions import (
     DataError,
     RedisError,
@@ -31,6 +31,7 @@ from redis.exceptions import (
     ExecAbortError,
     ReadOnlyError
 )
+from redis.selector import DefaultSelector
 from redis.utils import HIREDIS_AVAILABLE
 if HIREDIS_AVAILABLE:
     import hiredis
@@ -276,9 +277,7 @@ class PythonParser(BaseParser):
 
     def on_disconnect(self):
         "Called when the socket disconnects"
-        if self._sock is not None:
-            self._sock.close()
-            self._sock = None
+        self._sock = None
         if self._buffer is not None:
             self._buffer.close()
             self._buffer = None
@@ -498,6 +497,7 @@ class Connection(object):
             raise ConnectionError(self._error_message(e))
 
         self._sock = sock
+        self._selector = DefaultSelector(sock)
         try:
             self.on_connect()
         except RedisError:
@@ -529,7 +529,7 @@ class Connection(object):
                 if self.socket_keepalive:
                     sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
                     for k, v in iteritems(self.socket_keepalive_options):
-                        sock.setsockopt(socket.SOL_TCP, k, v)
+                        sock.setsockopt(socket.IPPROTO_TCP, k, v)
 
                 # set the socket_connect_timeout before we connect
                 sock.settimeout(self.socket_connect_timeout)
@@ -582,7 +582,8 @@ class Connection(object):
         if self._sock is None:
             return
         try:
-            self._sock.shutdown(socket.SHUT_RDWR)
+            if os.getpid() == self.pid:
+                self._sock.shutdown(socket.SHUT_RDWR)
             self._sock.close()
         except socket.error:
             pass
@@ -610,9 +611,9 @@ class Connection(object):
                 errmsg = e.args[1]
             raise ConnectionError("Error %s while writing to socket. %s." %
                                   (errno, errmsg))
-        except Exception as e:
+        except:  # noqa: E722
             self.disconnect()
-            raise e
+            raise
 
     def send_command(self, *args):
         "Pack and send a command to the Redis server"
@@ -624,16 +625,19 @@ class Connection(object):
         if not sock:
             self.connect()
             sock = self._sock
-        return self._parser.can_read() or \
-            bool(select([sock], [], [], timeout)[0])
+        return self._parser.can_read() or self._selector.can_read(timeout)
+
+    def is_ready_for_command(self):
+        "Check if the connection is ready for a command"
+        return self._selector.is_ready_for_command()
 
     def read_response(self):
         "Read the response from a previously sent command"
         try:
             response = self._parser.read_response()
-        except Exception as e:
+        except:  # noqa: E722
             self.disconnect()
-            raise e
+            raise
         if isinstance(response, ResponseError):
             raise response
         return response
@@ -975,7 +979,6 @@ class ConnectionPool(object):
                     # another thread already did the work while we waited
                     # on the lock.
                     return
-                self.disconnect()
                 self.reset()
 
     def get_connection(self, command_name, *keys, **options):
@@ -986,6 +989,23 @@ class ConnectionPool(object):
         except IndexError:
             connection = self.make_connection()
         self._in_use_connections.add(connection)
+        try:
+            # ensure this connection is connected to Redis
+            connection.connect()
+            # connections that the pool provides should be ready to send
+            # a command. if not, the connection was either returned to the
+            # pool before all data has been read or the socket has been
+            # closed. either way, reconnect and verify everything is good.
+            if not connection.is_ready_for_command():
+                connection.disconnect()
+                connection.connect()
+                if not connection.is_ready_for_command():
+                    raise ConnectionError('Connection not ready')
+        except:  # noqa: E722
+            # release the connection back to the pool so that we don't leak it
+            self.release(connection)
+            raise
+
         return connection
 
     def get_encoder(self):
@@ -1014,6 +1034,7 @@ class ConnectionPool(object):
 
     def disconnect(self):
         "Disconnects all connections in the pool"
+        self._checkpid()
         all_conns = chain(self._available_connections,
                           self._in_use_connections)
         for connection in all_conns:
@@ -1116,6 +1137,23 @@ class BlockingConnectionPool(ConnectionPool):
         if connection is None:
             connection = self.make_connection()
 
+        try:
+            # ensure this connection is connected to Redis
+            connection.connect()
+            # connections that the pool provides should be ready to send
+            # a command. if not, the connection was either returned to the
+            # pool before all data has been read or the socket has been
+            # closed. either way, reconnect and verify everything is good.
+            if not connection.is_ready_for_command():
+                connection.disconnect()
+                connection.connect()
+                if not connection.is_ready_for_command():
+                    raise ConnectionError('Connection not ready')
+        except:  # noqa: E722
+            # release the connection back to the pool so that we don't leak it
+            self.release(connection)
+            raise
+
         return connection
 
     def release(self, connection):
@@ -1135,5 +1173,6 @@ class BlockingConnectionPool(ConnectionPool):
 
     def disconnect(self):
         "Disconnects all connections in the pool."
+        self._checkpid()
         for connection in self._connections:
             connection.disconnect()
